// Copyright (c) 2005 David Drysdale.  See license.txt for license info
#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include <TxtGlue.h>
#include "resourcedefines.h"
#include "mainform.h"
#include "utils.h"
#include "constants.h"
#include "globals.h"
#include "gamestate.h"
#include "boardgadget.h"
#include "board.h"

static Boolean AppHandleEvent( EventPtr eventP)
{
  UInt16 formId;
  FormPtr form;
  
  if (eventP->eType == frmLoadEvent) {
    // Load the form resource.
    formId = eventP->data.frmLoad.formID;
    form = FrmInitForm(formId);
    ErrFatalDisplayIf(!form, "Can't initialize form");
    FrmSetActiveForm(form);
    
    // Set the event handler for the form.  The handler of the currently
    // active form is called by FrmHandleEvent each time is receives an
    // event.
    switch (formId) {
      case MainForm:
        FrmSetEventHandler(form, MainFormHandleEvent);
        break;

      case HelpDialog:
        FrmSetEventHandler(form, HelpEvent);
        break;

      default:
        ErrFatalDisplay("Invalid Form Load Event");
        break;
        
    }
    return true;
  } else if ((eventP->eType == nilEvent) && (FrmGetActiveForm() == FrmGetFormPtr(MainForm))) {
    if (GAMESTARTED(board) && !GAMEOVER(board)) {
      // Update the time taken 
      UInt32 timeNow = TimGetTicks();
      board->timeTaken += (timeNow - lastTimeCheck);
      lastTimeCheck = timeNow;
      // Redisplay it
      BoardShowNumbers(FrmGetFormPtr(MainForm));
    }
    return true;
  } else {
    return false;
  }
}

static void AppEventLoop(void)
{
  Err error;
  EventType event;

  do {
    EvtGetEvent(&event, kBlockTicks);

    // Don't let up or down key presses get through, so that the 
    // clicking sound doesn't get triggered.  This has the downside that 
    // the user can no longer use up/down to navigate the menus, but never mind. 
    if ((event.eType == keyDownEvent) && 
        !TxtCharIsHardKey(event.data.keyDown.modifiers, event.data.keyDown.chr) && 
        EvtKeydownIsVirtual(&event) && 
        ((event.data.keyDown.chr == vchrPageUp) || (event.data.keyDown.chr == vchrPageDown))) {
      // Do nothing
    } else if (! SysHandleEvent(&event))
      if (! MenuHandleEvent(0, &event, &error))
        if (! AppHandleEvent(&event))
          FrmDispatchEvent(&event);

  } while (event.eType != appStopEvent);
}

static Err AppStart(void)
{
  Boolean newGame;
  Err err = errNone;

  err = LoadGameState(&newGame);
  // If this is a new game or we've loaded a finished game, initialize the board
  if (newGame || GAMEOVER(board)) InitBoard(board);
  lastTimeCheck = TimGetTicks();

  // Start of day stuff for display code
  BoardGadgetInit();
  return err;
}

static void AppStop(void)
{
  SaveGameState();
  BoardGadgetFinit();
  FrmCloseAllForms();
}

UInt32 PilotMain(UInt16 launchCode, MemPtr launchParameters, 
                 UInt16 launchFlags)
{
  Err error;
  
  switch (launchCode) {
    case sysAppLaunchCmdNormalLaunch:
      error = RomVersionCompatible(kOurMinVersion, launchFlags);
      if (error) return error;
      error = AppStart();
      if (error) return error;
      
      FrmGotoForm(MainForm);
      AppEventLoop();
      AppStop();
      break;
      
    default:
      break;
  }
  
  return 0;
}

