// Copyright (c) 2005 David Drysdale.  See license.txt for license info
#define DO_NOT_ALLOW_ACCESS_TO_INTERNALS_OF_STRUCTS
#include <BuildDefines.h>
#ifdef DEBUG_BUILD
#define ERROR_CHECK_LEVEL ERROR_CHECK_FULL
#endif
#include <PalmOS.h>
#include "resourcedefines.h"
#include "constants.h"
#include "boardgadget.h"
#include "horizscroll.h"
#include "vertscroll.h"
#include "globals.h"
#include "utils.h"

void UpdateVScrollBar(FormPtr form, UInt16 scrollbarID, UInt16 currentPosition)
{
  ScrollBarPtr scroll;

  scroll = (ScrollBarPtr)FrmGetObjectPtr(form, FrmGetObjectIndex(form, scrollbarID));
  SclSetScrollBar(scroll, currentPosition, 0, MAX((board->height - kVertCells),0), kVertCells);
}

void UpdateHScrollBar(FormPtr form, UInt16 scrollbarID, UInt16 currentPosition)
{
  MemHandle data;
  HORIZSCROLL *pScrollData;
  UInt16 gadgetIndex;

  gadgetIndex = FrmGetObjectIndex(form, scrollbarID);
  assert(scrollbarID == HorizScroll);

  data = (MemHandle) FrmGetGadgetData(form, gadgetIndex);
  pScrollData = (HORIZSCROLL *)MemHandleLock(data);

  pScrollData->value = currentPosition;
  pScrollData->minValue = 0;
  pScrollData->maxValue = (board->width - kHorizCells);
  pScrollData->pageSize = kHorizCells;
  // Draw it
  HorizScrollDraw(form, HorizScroll);

  MemHandleUnlock(data);
}

Boolean ScrollHLines(FormPtr form, UInt16 scrollbarID, Int16 numToScroll, Boolean redraw)
{
  Int16 newHPos;

  newHPos = currHPos + numToScroll;
  if (newHPos < 0) return false;
  if (newHPos > (board->width - kHorizCells)) return false;

  UpdateHScrollBar(form, scrollbarID, newHPos);
  if (newHPos != currHPos) redraw = true;
  currHPos = newHPos;
  if (redraw) {
    UInt16 ii, jj;
    UInt16 firstCellToRedraw;
    RectangleType bounds;
    Coord newX;
    
    FrmGetObjectBounds(form, FrmGetObjectIndex(form, BoardGadget), &bounds);
    if (numToScroll < 0) {
      // Scroll up
      numToScroll = -numToScroll;
      firstCellToRedraw = 0;
      newX = bounds.topLeft.x + (numToScroll * kCellSize);
    } else {
      // Scroll down 
      firstCellToRedraw = kHorizCells - numToScroll;
      newX = bounds.topLeft.x;
      bounds.topLeft.x += (numToScroll * kCellSize);
    }
    bounds.extent.x = (kHorizCells - numToScroll) * kCellSize;
    bounds.extent.y = kVertCells * kCellSize;

    if (numToScroll >= kHorizCells) {
      // Scrolling more than the entire visible board
      firstCellToRedraw = 0;
      numToScroll = kHorizCells;
    } else {
      // Copy the parts of the screen that remain visible (if there are any)
      WinCopyRectangle(NULL, NULL, &bounds, newX, bounds.topLeft.y, winPaint);
    }

    // Now redraw the newly visible cells
    for (ii = firstCellToRedraw; ii < (firstCellToRedraw + numToScroll); ii++) {
      for (jj = 0; jj < kVertCells; jj++) {
        CellDraw(ii*kCellSize, jj*kCellSize, currHPos+ii, currVPos+jj);
      }
    }
  }

  return redraw;
}
  
void ScrollVLines(FormPtr form, UInt16 scrollbarID, Int16 numToScroll, Boolean redraw)
{
  Int16 newVPos;

  newVPos = currVPos + numToScroll;
  assert(newVPos >= 0);
  assert(newVPos <= (board->height - kVertCells));

  UpdateVScrollBar(form, scrollbarID, newVPos);
  if (newVPos != currVPos) redraw = true;
  currVPos = newVPos;
  if (redraw) {
    UInt16 ii, jj;
    UInt16 firstCellToRedraw;
    RectangleType bounds;
    Coord newY;
    
    FrmGetObjectBounds(form, FrmGetObjectIndex(form, BoardGadget), &bounds);
    if (numToScroll < 0) {
      // Scroll up
      numToScroll = -numToScroll;
      firstCellToRedraw = 0;
      newY = bounds.topLeft.y + (numToScroll * kCellSize);
    } else {
      // Scroll down 
      firstCellToRedraw = kVertCells - numToScroll;
      newY = bounds.topLeft.y;
      bounds.topLeft.y += (numToScroll * kCellSize);
    }
    bounds.extent.x = kHorizCells * kCellSize;
    bounds.extent.y = (kVertCells - numToScroll) * kCellSize;
    
    if (numToScroll >= kVertCells) {
      // Scrolling more than the entire visible board
      firstCellToRedraw = 0;
      numToScroll = kVertCells;
    } else {
      // Copy the parts of the screen that remain visible (if there are any)
      WinCopyRectangle(NULL, NULL, &bounds, bounds.topLeft.x, newY, winPaint);
    }

    // Now redraw the newly visible cells
    for (ii=0; ii<kHorizCells; ii++) {
      for (jj = firstCellToRedraw; jj < (firstCellToRedraw + numToScroll); jj++) {
        CellDraw(ii*kCellSize, jj*kCellSize, currHPos+ii, currVPos+jj);
      }
    }
  }
}

void ScrollHPage(FormPtr form, UInt16 scrollbarID, WinDirectionType direction)
{
  Int16 numToScroll;
  
  if (direction == winUp) {
    if (currHPos < kHorizCells) {
      numToScroll = -currHPos;
    } else {
      numToScroll = -(kHorizCells - 1);
    }
  } else {
    if (currHPos + kHorizCells - 1 > (board->width - kHorizCells)) {
      numToScroll = (board->width - kHorizCells - currHPos);
    } else {
      numToScroll = (kHorizCells - 1);
    }
  }
  (void)ScrollHLines(form, scrollbarID, numToScroll, true);
}

void ScrollVPage(FormPtr form, UInt16 scrollbarID, WinDirectionType direction)
{
  Int16 numToScroll;
  
  if (direction == winUp) {
    if (currVPos < kVertCells) {
      numToScroll = -currVPos;
    } else {
      numToScroll = -(kVertCells - 1);
    }
  } else {
    if (currVPos + kVertCells - 1 > (board->height - kVertCells)) {
      numToScroll = (board->height - kVertCells - currVPos);
    } else {
      numToScroll = (kVertCells - 1);
    }
  }
  ScrollVLines(form, scrollbarID, numToScroll, true);
}

void ScrollToKeepCellVisible(UInt16 x, UInt16 y)
{
  FormPtr form = FrmGetActiveForm();
  if (x < currHPos) ScrollHLines(form, HorizScroll, (x - currHPos), true);
  if (x >= (currHPos + kHorizCells)) ScrollHLines(form, HorizScroll, x + 1 - currHPos - kHorizCells, true);
  if (y < currVPos) ScrollVLines(form, VertScroll, (y - currVPos), true);
  if (y >= (currVPos + kVertCells)) ScrollVLines(form, VertScroll, y + 1 - currVPos - kVertCells, true);
}

void HorizScrollDraw(FormPtr form, UInt16 gadgetID)
{
  MemHandle data;
  HORIZSCROLL *pScrollData;
  RectangleType bounds;
  RectangleType line;
  RectangleType tabBounds;
  UInt16 gadgetIndex;
  UInt16 ii;
  UInt16 midpoint;
  Int16 delta;
  Coord x;
  static const CustomPatternType pattern = {0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55, 0xAA, 0x55};

  gadgetIndex = FrmGetObjectIndex(form, gadgetID);
  assert(gadgetID == HorizScroll);
  FrmGetObjectBounds(form, gadgetIndex, &bounds);

  data = (MemHandle) FrmGetGadgetData(form, gadgetIndex);
  pScrollData = (HORIZSCROLL *)MemHandleLock(data);

  // If there is no range to the scrollbar, give up and clear the area
  if (pScrollData->maxValue <= pScrollData->minValue) {
    WinEraseRectangle(&bounds, 0);
    return;
  }

  // Draw the triangles
  tabBounds.topLeft.y = bounds.topLeft.y;
  tabBounds.extent.x = kScrollTabWidth;
  tabBounds.extent.y = bounds.extent.y;
  tabBounds.topLeft.x = bounds.topLeft.x;
  WinEraseRectangle(&tabBounds, 0);
  tabBounds.topLeft.x = bounds.topLeft.x + bounds.extent.x - kScrollTabWidth;
  WinEraseRectangle(&tabBounds, 0);
  midpoint = bounds.topLeft.y + (bounds.extent.y / 2);
  for (ii = 1; ii < (kScrollTabWidth - 1); ii++) {
    x = bounds.topLeft.x + ii;
    delta = ii - 1;
    WinDrawLine(x, midpoint - delta, x, midpoint + delta);
    x = bounds.topLeft.x + bounds.extent.x - 1 - ii;
    WinDrawLine(x, midpoint - delta, x, midpoint + delta);
  }
 
  // Draw gray line
  line.topLeft.x = bounds.topLeft.x + kScrollTabWidth;
  line.topLeft.y = midpoint - (kScrollLineWidth / 2);
  line.extent.x = bounds.extent.x - (2*kScrollTabWidth);
  line.extent.y = kScrollLineWidth;
  WinSetPattern(&pattern);
  WinFillRectangle(&line, 0);

  // Draw solid line
  line.topLeft.x += ((pScrollData->value * line.extent.x) / 
                     (pScrollData->maxValue + pScrollData->pageSize));
  line.extent.x *= pScrollData->pageSize;
  line.extent.x /= (pScrollData->maxValue + pScrollData->pageSize - pScrollData->minValue);
  line.extent.x++;

  // Remember where this is for the tap processing
  pScrollData->leftCoord = line.topLeft.x;
  pScrollData->xExtent = line.extent.x;
  WinDrawRectangle(&line, 0);
  
  // Erase slightly to left and right
  WinEraseLine(line.topLeft.x - 1, midpoint - 1, 
               line.topLeft.x - 1, midpoint + 1);
  WinEraseLine(line.topLeft.x + line.extent.x, midpoint - 1, 
               line.topLeft.x + line.extent.x, midpoint + 1);

  MemHandleUnlock(data);
}


void HorizScrollTap(FormPtr form, UInt16 gadgetID, EventPtr event)
{
  MemHandle data;
  HORIZSCROLL *pScroll;
  RectangleType bounds;
  UInt16 gadgetIndex;
  Int16 xRel;
  RectangleType tabBounds;
  Int16 direction = 1;
  Int16 x, y;
  Boolean penDown;

  gadgetIndex = FrmGetObjectIndex(form, gadgetID);
  assert(gadgetID == HorizScroll);
  FrmGetObjectBounds(form, gadgetIndex, &bounds);

  data = (MemHandle) FrmGetGadgetData(form, gadgetIndex);
  pScroll = (HORIZSCROLL *)MemHandleLock(data);

  // There are five possible sections for the user to click
  // on:                 bounds.topLeft.x + ..
  //  - left hand tab    [0, kScrollTabWidth)
  //  - left gray line   [kScrollTabWidth, pScroll->leftCoord)
  //  - solid line       [pScroll->leftCoord, pScroll->leftCoord + pScroll->xExtent)
  //  - right gray line  [pScroll->leftCoord + pScroll->xExtent, bounds.extent.x - kScrollTabWidth)
  //  - right hand tab   [bounds.extent.x - kScrollTabWidth, bounds.extent.x)
  
  xRel = event->screenX - bounds.topLeft.x;
  if ((xRel < kScrollTabWidth) || (xRel >= (bounds.extent.x - kScrollTabWidth))) {
    // Left and right tabs
    if (xRel < kScrollTabWidth) {
      direction = -1;
      tabBounds.topLeft.x = bounds.topLeft.x;
    } else if (xRel >= (bounds.extent.x - kScrollTabWidth)) {
      direction = 1;
      tabBounds.topLeft.x = bounds.topLeft.x + bounds.extent.x - kScrollTabWidth;
    }
    tabBounds.topLeft.y = bounds.topLeft.y;
    tabBounds.extent.x = kScrollTabWidth;
    tabBounds.extent.y = bounds.extent.y;

    // Do highlighting until pen up
    WinInvertRectangle(&tabBounds, 0);
    do {
      if (ScrollHLines(form, gadgetID, direction, true))
        WinInvertRectangle(&tabBounds, 0);
      PenGetPoint(&x, &y, &penDown);
    } while (penDown);
    // Flip back again
    WinInvertRectangle(&tabBounds, 0);
  } else if (xRel < pScroll->leftCoord) {
    // Left page scroll area
    do {
      if (xRel < pScroll->leftCoord) 
        ScrollHPage(form, gadgetID, winUp);
      PenGetPoint(&x, &y, &penDown);
      xRel = x - bounds.topLeft.x;
    } while (penDown);
  } else if (xRel < (pScroll->leftCoord + pScroll->xExtent)) {
    Int16 lastScrollX = xRel;
    Int16 scrollTick = ((bounds.extent.x - (2*kScrollTabWidth)) / 
                        (pScroll->maxValue + pScroll->pageSize));
    // Slider itself
    do {
      if (xRel < (lastScrollX - scrollTick)) {
        lastScrollX = xRel;
        ScrollHLines(form, gadgetID, -1, true);
      } else if (xRel >= (lastScrollX + scrollTick)) {
        lastScrollX = xRel;
        ScrollHLines(form, gadgetID, 1, true);
      }

      PenGetPoint(&x, &y, &penDown);
      xRel = x - bounds.topLeft.x;
    } while (penDown);
  } else {
    // Right page scroll area
    do {
      if (xRel >= (pScroll->leftCoord + pScroll->xExtent))
        ScrollHPage(form, gadgetID, winDown);
      PenGetPoint(&x, &y, &penDown);
      xRel = x - bounds.topLeft.x;
    } while (penDown);
  }

  MemHandleUnlock(data);
}

